//************************************************************************
//*                                                                      *
//*   DOS driver for esd ISA-331 CAN-Interface Board                     *
//*                                                                      *
//*----------------------------------------------------------------------*
//*                                                                      *
//*      Filename:       CAN.H                                           *
//*      Date:           23-Aug-97                                       *
//*      Language:       Borland C++ V 3.1                               *
//*      Dev.system:     IBM PC  / DOS                                   *
//*                                                                      *
//*      Purpose:        Low Level Driver Routines for esd ISA-331       *
//*                      with Firmware upgrades, this Routines may       *
//*                      change, only user API as given in CAN.H         *
//*                      will not change.                                *
//*----------------------------------------------------------------------*
//* Rev. | Released | Programmer  | Comments                             *
//*----------------------------------------------------------------------*
//* 1.00 | 28.05.97 | Ar          | Initial Release                      *
//* 1.10 | 30.06.97 | Ar          | For CAN 2.0B                         *
//* 1.11 | 23.08.97 | Ar          | Bug Fix: Copy Error Code in CanRead  *
//* 1.12 | 18.11.97 | Ta          | PC104                            *
//************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#ifdef __DJGPP__
#include <dpmi.h>
#endif
#include "can.h"

#ifdef __cplusplus
    #define __CPPARGS ...
#else
    #define __CPPARGS
#endif

#define INTBASE1 0x20
#define INTBASE2 0xA0
#define IMSK     1
#define OCW2     0
#define EndOfInt 0x20

#define swap(data) (((data << 8) & 0xFF00) | ((data >> 8) & 0x00FF))

#ifdef __DJGPP__
#define CANBUFF 0x10000
#else
#define CANBUFF 0x7000
#endif
#define MAX_CAN_NET 2

//*----------------------------------------------------------------------*
//*      P R O C E D U R E S
//*----------------------------------------------------------------------*
typedef struct
    {
    volatile unsigned char buffer[CANBUFF];
    volatile          int   bufferIn;
    volatile          int   bufferOu;
    volatile          int   msgLostErr;
    volatile          int   msgFrameErr;
    } CANBUF;


typedef struct
    {
    int canIoPort;
    int canIrqMask;
    int canIrqVect;
    int canIrqLevel;
    int canIrqCnt;
    volatile int   intBufferPtr;
    volatile short intBuffer[20];
    } ISA331;

#ifdef __DJGPP__
#define interrupt
_go32_dpmi_seginfo canOldVector;
_go32_dpmi_seginfo canNewVector;
#else
void interrupt (*oldhandler)(__CPPARGS);
#endif

ISA331 isa331;
CANBUF canBuf;

CAN_MSG msx,msy;


void interrupt canIrq(__CPPARGS)
    {
    int mdata,l,l1;

    while ((inp(isa331.canIoPort+1) & 0x10) == 0x10)
	{
	mdata=inpw(isa331.canIoPort);
	if (mdata & 0x8000)
	    {
	    if (isa331.intBufferPtr != 0)
		canBuf.msgFrameErr++;
	    isa331.intBufferPtr = 0;
	    }
	isa331.intBuffer[isa331.intBufferPtr++] = mdata;
	if (isa331.intBufferPtr == 20)
	    {
	    isa331.intBufferPtr = 0;
	    l = canBuf.bufferIn;
	    for (l1=0;l1<20;l1++)
		{
		canBuf.buffer[l++] =(unsigned char) (isa331.intBuffer[l1] & 0x00ff);
		if (l == CANBUFF)
		    l = 0;
		if (l == canBuf.bufferOu)
		    {
		    canBuf.msgLostErr++;
		    break;
		    }
		}
	    if (l1 == 20)
		canBuf.bufferIn = l;
	    }
	}
    if (isa331.canIrqLevel >= 8)
	{
	outp(INTBASE2+OCW2,EndOfInt);
	}
    outp(INTBASE1+OCW2,EndOfInt);
    isa331.canIrqCnt++;
    }


int canEnableIrq
    (
    int            level
    )
    {
    int l,l1,stat,mdata;
#ifdef __DJGPP__
    int status;
#endif
    isa331.intBufferPtr = 0;
    isa331.canIrqLevel  = level ;
    isa331.canIrqCnt    = 0;
#ifdef __DJGPP__
    status = _go32_dpmi_lock_data(&isa331      ,sizeof(isa331));
    if (status != 0)
	{
        printf("Can not lock pci331 info\n");
        return(-1);
	}
    status = _go32_dpmi_lock_data(&canBuf      ,sizeof(canBuf));
    if (status != 0)
        {
        printf("Can not lock canBuffer  \n");
	return(-1);
        }
#endif
    disable();
    
#ifdef PC104
    outp(isa331.canIoPort+0x6,0x06);    // Board Reset
#else
    outp(isa331.canIoPort+0x6,0x86);    // Board Reset
#endif

    sleep(1);
    while ((inp(isa331.canIoPort+1) & 0x10) == 0x10)
        {
	int mdata;
	inpw(isa331.canIoPort);
        }
    l1       = level;
    for (l=2;l<6;l++)
        {
        stat=l;
	if (l1 & 0x01)
            {
            stat |= 0x80;
            }
        outp(isa331.canIoPort+6,stat);
	l1=l1>>1;
        }
    outp(isa331.canIoPort+6,0x87);     // Enable IRQ Pin
    outp(isa331.canIoPort+6,0x81);     // Enable IRQ

    if(level < 8)
	{
        isa331.canIrqMask  = 1 << level;
	isa331.canIrqVect = 8 +  level;
	}
    else
        {
        isa331.canIrqMask  = 1 <<  (level-8);
        isa331.canIrqVect = 0x70 + level-8;
        }
    if(isa331.canIrqLevel >= 8)
        {
        outp(INTBASE2+IMSK, inp(INTBASE2+IMSK) & ~isa331.canIrqMask); // freigeben
        }
    else
	{
        outp(INTBASE1+IMSK, inp(INTBASE1+IMSK) & ~isa331.canIrqMask); // freigeben
	}
#ifdef __DJGPP__
    canNewVector.pm_selector = _my_cs();
    canNewVector.pm_offset   = (int)canIrq;
    status = _go32_dpmi_allocate_iret_wrapper(&canNewVector);
    status = _go32_dpmi_set_protected_mode_interrupt_vector(isa331.canIrqVect,&canNewVector);
#else
    oldhandler = _dos_getvect(isa331.canIrqVect);
    _dos_setvect(isa331.canIrqVect, &canIrq);
#endif

#ifdef PC104
    outp(isa331.canIoPort+6,0x86);     // Remove Board Reset
#else
    outp(isa331.canIoPort+6,0x06);     // Remove Board Reset
#endif

    enable();
    return(0);
    }


int canDisableIrq
    (
    void
    )
    {
    int l;
#ifdef __DJGPP__
    int status;
#endif

    disable();
    if(isa331.canIrqLevel >= 8)
        {
        outp(INTBASE2+IMSK, inp(INTBASE2+IMSK) | isa331.canIrqMask); // freigeben
        }
    else
        {
	outp(INTBASE1+IMSK, inp(INTBASE1+IMSK) | isa331.canIrqMask); // freigeben
        }
#ifdef __DJGPP__
    status = _go32_dpmi_set_protected_mode_interrupt_vector(isa331.canIrqVect,&canOldVector);
    status = _go32_dpmi_free_iret_wrapper(&canNewVector);
    printf("There were %ld IRQ's\n",isa331.canIrqCnt);
#else
    _dos_setvect(isa331.canIrqVect, oldhandler);
#endif
    outp(isa331.canIoPort+6,0x07);       // Disable EN2
    outp(isa331.canIoPort+6,0x00);       // Disable EN0
    outp(isa331.canIoPort+6,0x01);       // Disable EN1
    for (l=2;l<6;l++)
    outp(isa331.canIoPort+6,l);      // Back to level 0
    enable();
    return(0);
    }


void canOut
    (
    unsigned short data
    )
    {
    while ((inp(isa331.canIoPort+1) & 0x40) == 0x00)
	{
        }
    outpw(isa331.canIoPort,data);
    }


int canCmmd
    (
    CAN_MSG* cmmd
    )
    {
    int l;
    unsigned char* data;

    data = (unsigned char*) cmmd;
    canOut(0x8000 | ((*data++) & 0x0ff));
    for (l=1;l<20;l++)
        {
        canOut((*data++) & 0x0ff);
        }
    return(0);
    }


int canMsg
    (
    CAN_MSG* msg
    )
    {
    int             stat;
    int             l,l1;
    unsigned char*  dptr;
    unsigned short  mdata;


    if (isa331.canIrqLevel == 0)
        {
        stat = (inp(isa331.canIoPort+1) & 0x10);
        if (stat == 0)
            {
	    return(-1);
            }
	memset(msg,0,sizeof(CAN_MSG));
	dptr =(unsigned char *) msg;
	for (l=0;l<20;l++)
	    {
	    while ((inp(isa331.canIoPort+1) & 0x10) == 0);
		mdata=inpw(isa331.canIoPort);
	    if ((mdata & 0x8000) != 0)
		{
		if ( l != 0)
		    {
		    printf("%04X  ",mdata);
		    dptr =(unsigned char *) msg;
		    l=0;
		    }
		}
	    *dptr++=mdata & 0x0ff;
	    }
	}
    else
	{
	if (canBuf.bufferIn == canBuf.bufferOu)
	    {
	    return(-1);
	    }
	memset(msg,0,sizeof(CAN_MSG));
	dptr =(unsigned char *) msg;
	for (l=0;l<20;l++)
	    {
	    while (canBuf.bufferOu == canBuf.bufferIn);
		mdata = canBuf.buffer[canBuf.bufferOu];
	    l1 = canBuf.bufferOu + 1;
	    if (l1 == CANBUFF)
		l1 = 0;
	    canBuf.bufferOu = l1;
//	    if ((mdata & 0x8000) != 0)
//		{
//		if ( l != 0)
//		    {
//		    printf("Out of sync data %04X %04x %04x %04x\n",
//			mdata,canBuf.bufferOu,canBuf.bufferIn,canBuf.lost);
//		    dptr =(unsigned char *) msg;
//		    l=0;
//		    }
//		}
	    *dptr++=mdata & 0x0ff;
	    }
	}
    return(0);
    }


int canHwInit
    (
    int port
    )
    {
    if ((port < 0) || (port > 0x3f8))
	return(-1);
    isa331.canIoPort   = port;
    isa331.canIrqLevel = 0;
    canBuf.bufferIn    = 0;
    canBuf.bufferOu    = 0;
    canBuf.msgLostErr  = 0;
    canBuf.msgFrameErr = 0;
    return(0);
    }


int canHwReset
    (
    void
    )
    {
#ifdef PC104
    outp(isa331.canIoPort+0x6,0x06);    // Board Reset
#else
    outp(isa331.canIoPort+0x6,0x86);    // Board Reset
#endif
    sleep(1);
    while ((inp(isa331.canIoPort+1) & 0x10) == 0x10)
	{
	inpw(isa331.canIoPort);         // Clear FIFO
	}

#ifdef PC104
    outp(isa331.canIoPort+6,0x86);      // Remove Board Reset
#else
    outp(isa331.canIoPort+6,0x06);      // Remove Board Reset
#endif
    return(0);
    }


int canUpdateFirmware
    (
    unsigned char* key,
    unsigned char* firmware,
    unsigned long  length,
    unsigned long  block,
    unsigned long  blocksize
    )
    {
    CAN_MSG msg;
    int     status;
    int     l,l1;

    memset(&msg,0,sizeof(msg));
    msg.cmmd=CAN_HM_FLASH;
    memcpy(&msg.net,key,18);
    status=canCmmd(&msg);

    memset(&msg,0,sizeof(msg));
    msg.cmmd=CAN_HM_FERASE;
    msg.data[3] = (unsigned char) (block & 0x0ff); block = block >> 8;
    msg.data[2] = (unsigned char) (block & 0x0ff); block = block >> 8;
    msg.data[1] = (unsigned char) (block & 0x0ff); block = block >> 8;
    msg.data[0] = (unsigned char) (block & 0x0ff);
    status=canCmmd(&msg);
    for(l=0;l<length;l +=14)
        {
        memset(&msg,0,sizeof(msg));
        msg.cmmd=CAN_HM_FPROG;
        msg.id  = swap(l);
        l1      = length - l;
        if (l1 >14)
            l1 = 14;
        msg.len = swap(l1);
        memcpy(&msg.data[0],&firmware[l],l1);
        status=canCmmd(&msg);
        }
    memset(&msg,0,sizeof(msg));
    msg.cmmd=CAN_HM_FEXIT;
    status=canCmmd(&msg);
    return(0);
    }


int canVerifyFirmware
    (
    unsigned int   offset,
    unsigned int   length,
    unsigned char* firmware
    )
    {
    CAN_MSG request,reply;
    memset(&request,0,sizeof(request));
    request.net  = (offset >> 16) & 0x0ff;
    request.id   = swap((offset & 0x0ffff));
    request.len  = swap(length);
    request.cmmd = CAN_HM_FPVERIFY;

    return(0);
    }

int canInit
    (
    int net,
    int baudrate
    )
    {
    CAN_MSG msg;

    if ((net <0) || (net > (MAX_CAN_NET-1)))
        return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_BAUD;
    msg.net     = net;
    msg.data[1] = baudrate;
    return(canCmmd(&msg));
    }


int canInitEx
    (
    int net,
    int bt0_bt1
    )
    {
    CAN_MSG msg;

    if ((net <0) || (net > (MAX_CAN_NET-1)))
        return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_EX_BAUD;
    msg.net     = net;
    msg.data[0] = bt0_bt1 / 0x100;
    msg.data[1] = bt0_bt1 % 0x100;
    return(canCmmd(&msg));
    }


int canRead
    (
    int*           net,
    int*           id,
    int*           length,
    unsigned char* data
    )
    {
    int            l;
    int            stat;
    CAN_MSG        msg;

    if (canMsg(&msg) != 0)
        return(-1);
    *net    = msg.net;
    *length = swap(msg.len);
    *id     = swap(msg.id ) & 0x7ff;
    for (l=0;l < 8;l++)
        {
        *data++ = msg.data[l];
        }
    return(msg.cmmd);
    }


int canReadEx
    (
    int*           net,
    long*          id,
    int*           length,
    unsigned char* data
    )
    {
    int            l;
    int            stat;
    CAN_MSG        msg;

    if (canMsg(&msg) != 0)
        return(-1);
    *net    = msg.net;
    *length = swap(msg.len);
    *id     = swap(msg.id );
    if ((msg.cmmd == CAN_HM_EX_BCAN)  ||
	(msg.cmmd == CAN_HM_EX_TXDONE)||
        (msg.cmmd == CAN_HM_EX_TXTOUT))
        {
        *id = (*id) * 0x10000 + swap(msg.x2);
        }
    if ((*length > 0) && (*length <9))
        {
        for (l=0;l < *length;l++)
            {
            *data++ = msg.data[l];
            }
        }
    return(msg.cmmd);
    }


int canWrite
    (
    int            net,
    int            id,
    int            length,
    unsigned char* data
    )
    {
    CAN_MSG msg;
    int     l;

    if ((net <0) || (net > (MAX_CAN_NET-1)))
        return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_BCAN;
    msg.net     = net;
    msg.id      = swap(id);
    msg.len     = swap(length);
    if ((length>0) && (length <9))
        {
        for (l=0;l<length;l++)
            msg.data[l]=data[l];
        }
    memcpy(&msx,&msg,sizeof(msg));
    return(canCmmd(&msg));
    }


int canWriteEx
    (
    int            net,
    long           id,
    int            length,
    unsigned char* data
    )
    {
    CAN_MSG msg;
    int     l;

    if ((net <0) || (net > (MAX_CAN_NET-1)))
        return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_EX_BCAN;
    msg.net     = net;
    msg.id      = swap(id / 0x10000);
    msg.x2      = swap(id % 0x10000);
    msg.len     = swap(length);
    if ((length>0) && (length <9))
        {
        for (l=0;l<length;l++)
            msg.data[l]=data[l];
        }
    return(canCmmd(&msg));
    }


int canEnableId
    (
    int net,
    int id
    )
    {
    CAN_MSG msg;

    if ((id < 0) || (id > 2047))
        return(-1);
    if ((net <0) || (net > (MAX_CAN_NET-1)))
        return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_ENABLE;
    msg.net     = net;
    msg.id      = swap(id);
    return(canCmmd(&msg));
    }


int canDisableId
    (
    int net,
    int id
    )
    {
    CAN_MSG msg;

    if ((id < 0) || (id > 2047))
        return(-1);
    if ((net <0) || (net > (MAX_CAN_NET-1)))
        return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_DISABLE;
    msg.net     = net;
    msg.id      = swap(id);
    return(canCmmd(&msg));
    }


int canSetAcceptanceEx
    (
    int  net,
    int  handle,
    long mask,
    long id
    )
    {
    CAN_MSG msg;

    if ((net <0) || (net > (MAX_CAN_NET-1)))
        return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_EX_FILTER;
    msg.net     = net;
    msg.id      = swap(handle);
    msg.data[0] = (id   >> 24) & 0x0ff;
    msg.data[1] = (id   >> 16) & 0x0ff;
    msg.data[2] = (id   >>  8) & 0x0ff;
    msg.data[3] = (id        ) & 0x0ff;
    msg.data[4] = (mask >> 24) & 0x0ff;
    msg.data[5] = (mask >> 16) & 0x0ff;
    msg.data[6] = (mask >>  8) & 0x0ff;
    msg.data[7] = (mask      ) & 0x0ff;
    return(canCmmd(&msg));
    }

