//************************************************************************
//*                                                                      *
//*   DOS driver for esd ISA-331 CAN-Interface Board                     *
//*                                                                      *
//*----------------------------------------------------------------------*
//*                                                                      *
//*      Filename:       CAN.H                                           *
//*      Date:           28-MAY-97                                       *
//*      Language:       Borland C++ V 3.1                               *
//*      Dev.system:     IBM PC  / DOS                                   *
//*                                                                      *
//*      Purpose:        Low Level Driver Routines for esd ISA-331       *
//*                      with Firmware upgrades, this Routines may       *
//*                      change, only user API as given in CAN.H         *
//*                      will not change.                                *
//*----------------------------------------------------------------------*
//* Rev. | Released | Programmer  | Comments                             *
//*----------------------------------------------------------------------*
//* 1.00 | 28.05.97 | Ar          | Initial Release                      *
//************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include "can.h"

#ifdef __cplusplus
    #define __CPPARGS ...
#else
    #define __CPPARGS
#endif

#define INTBASE1 0x20
#define INTBASE2 0xA0
#define IMSK     1
#define OCW2     0
#define EndOfInt 0x20

#define swap(data) (((data << 8) & 0xFF00) | ((data >> 8) & 0x00FF))

#define CANBUFF 0x4000
#define MAX_CAN_NET 1

//*----------------------------------------------------------------------*
//*      P R O C E D U R E S
//*----------------------------------------------------------------------*

volatile short canBuffer[CANBUFF];
volatile int   canBufferIn = 0x0;
volatile int   canBufferOu = 0x0;

int            canIoPort   = 0x1e0;

int            canIrqMsk;
int            canIrqVect;
int            canIrqLevel;

void interrupt (*oldhandler)(__CPPARGS);
// void interrupt (*oldhandler)(void);

void interrupt canIrq(__CPPARGS)
    {
    int mdata,l;

    while ((inp(canIoPort+1) & 0x10) == 0x10)
	{
	mdata=inpw(canIoPort);
	canBuffer[canBufferIn]=mdata;
	l=canBufferIn+1;
	if (l == CANBUFF)
	    l = 0;
	if (l != canBufferOu)
	    canBufferIn = l;
	}
    if (canIrqLevel >= 8)
	{
	outp(INTBASE2+OCW2,EndOfInt);
	}
    outp(INTBASE1+OCW2,EndOfInt);
    }


int canEnableIrq
    (
    int            level
    )
    {
    int l,l1,stat,mdata;

    canIrqLevel = level ;

    disable();
    outp(canIoPort+0x6,0x86);    // Board Reset
    sleep(1);
    while ((inp(canIoPort+1) & 0x10) == 0x10)
	{
	int mdata;
	inpw(canIoPort);
	}
    l1       = level;
    for (l=2;l<6;l++)
	{
	stat=l;
	if (l1 & 0x01)
	    {
	    stat |= 0x80;
	    }
	outp(canIoPort+6,stat);
	l1=l1>>1;
	}
    outp(canIoPort+6,0x87);     // Enable IRQ Pin
    outp(canIoPort+6,0x81);     // Enable IRQ

    if(level < 8)
	{
	canIrqMsk  = 1 << level;
	canIrqVect = 8 +  level;
	}
    else
	{
	canIrqMsk  = 1 <<  (level-8);
	canIrqVect = 0x70 + level-8;
	}
    if(canIrqLevel >= 8)
	{
	outp(INTBASE2+IMSK, inp(INTBASE2+IMSK) & ~canIrqMsk); // freigeben
	}
    else
	{
	outp(INTBASE1+IMSK, inp(INTBASE1+IMSK) & ~canIrqMsk); // freigeben
	}
    oldhandler = _dos_getvect(canIrqVect);
    _dos_setvect(canIrqVect, &canIrq);

    outp(canIoPort+6,0x06);     // Remove Board Reset
    enable();
    return(0);
    }


int canDisableIrq
    (
    void
    )
    {
    int l;
    disable();
    if(canIrqLevel >= 8)
	{
	outp(INTBASE2+IMSK, inp(INTBASE2+IMSK) | canIrqMsk); // freigeben
	}
    else
	{
	outp(INTBASE1+IMSK, inp(INTBASE1+IMSK) | canIrqMsk); // freigeben
	}
    _dos_setvect(canIrqVect, oldhandler);
    outp(canIoPort+6,0x07);       // Disable EN2
    outp(canIoPort+6,0x00);       // Disable EN0
    outp(canIoPort+6,0x01);       // Disable EN1
    for (l=2;l<6;l++)
	outp(canIoPort+6,l);      // Back to level 0
    enable();
    return(0);
    }


void canOut
    (
    unsigned short data
    )
    {
    while ((inp(canIoPort+1) & 0x40) == 0x00)
	{
	}
    outpw(canIoPort,data);
    }


int canCmmd
    (
    CAN_MSG* cmmd
    )
    {
    int l;
    unsigned char* data;

    data = (unsigned char*) cmmd;
    canOut(0x8000 | ((*data++) & 0x0ff));
    for (l=1;l<20;l++)
	canOut((*data++) & 0x0ff);
    return(0);
    }


int canMsg
    (
    CAN_MSG* msg
    )
    {
    int             stat;
    int             l,l1;
    unsigned char*  dptr;
    unsigned short  mdata;


    if (canIrqLevel == 0)
	{
	stat = (inp(canIoPort+1) & 0x10);
	if (stat == 0)
	    {
	    return(-1);
	    }
	memset(msg,0,sizeof(CAN_MSG));
	dptr =(unsigned char *) msg;
	for (l=0;l<20;l++)
	    {
	    while ((inp(canIoPort+1) & 0x10) == 0);
	    mdata=inpw(canIoPort);
	    if ((mdata & 0x8000) != 0)
		{
		if ( l != 0)
		    {
		    printf("%04X  ",mdata);
		    dptr =(unsigned char *) msg;
		    l=0;
		    }
		}
	    *dptr++=mdata & 0x0ff;
//	    printf("%02X ", mdata &0x0ff);
	    }
//	printf("\n");
	}
    else
	{
	if (canBufferIn == canBufferOu)
	    {
	    return(-1);
	    }
	memset(msg,0,sizeof(CAN_MSG));
	dptr =(unsigned char *) msg;
	for (l=0;l<20;l++)
	    {
	    while (canBufferOu == canBufferIn);
	    mdata = canBuffer[canBufferOu];
	    l1 = canBufferOu+1;
	    if (l1 == CANBUFF)
		l1 = 0;
	    canBufferOu = l1;
	    if ((mdata & 0x8000) != 0)
		{
		if ( l != 0)
		    {
		    printf("%04X  ",mdata);
		    dptr =(unsigned char *) msg;
		    l=0;
		    }
		}
	    *dptr++=mdata & 0x0ff;
//	    printf("%02X ", mdata &0x0ff);
	    }
//	printf("\n");
	}
    return(0);
    }


int canHwInit
    (
    int port
    )
    {
    if ((port < 0) || (port > 0x3f8))
	return(-1);
    canIoPort   = port;
    canIrqLevel = 0;
    return(0);
    }


int canHwReset
    (
    void
    )
    {
    outp(canIoPort+0x6,0x86);    // Board Reset
    sleep(1);
    while ((inp(canIoPort+1) & 0x10) == 0x10)
	{
	inpw(canIoPort);         // Clear FIFO
	}
    outp(canIoPort+6,0x06);      // Remove Board Reset
    return(0);
    }


int canInit
    (
    int net,
    int baudrate
    )
    {
    CAN_MSG msg;

    if ((net <0) || (net > (MAX_CAN_NET-1)))
	return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = 4;
    msg.net     = net;
    msg.data[1] = baudrate;
    return(canCmmd(&msg));
    }


int canRead
    (
    int*           net,
    int*           id,
    int*           length,
    unsigned char* data
    )
    {
    int            l;
    int            stat;
    CAN_MSG        msg;

    if (canMsg(&msg) != 0)
	return(-1);
    *net    = msg.net;
    *length = swap(msg.len);
    *id     = swap(msg.id ) & 0x7ff;
    if ((*length > 0) && (*length <9))
	{
	for (l=0;l < *length;l++)
	    {
	    *data++ = msg.data[l];
	    }
	}
    return(msg.cmmd);
    }


int canWrite
    (
    int            net,
    int            id,
    int            length,
    unsigned char* data
    )
    {
    CAN_MSG msg;
    int     l;

    if ((net <0) || (net > (MAX_CAN_NET-1)))
	return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_BCAN;
    msg.net     = net;
    msg.id      = swap(id);
    msg.len     = swap(length);
    if ((length>0) && (length <9))
	{
	for (l=0;l<length;l++)
	    msg.data[l]=data[l];
	}
    return(canCmmd(&msg));
    }


int canEnableId
    (
    int net,
    int id
    )
    {
    CAN_MSG msg;

    if ((id < 0) || (id > 2047))
	return(-1);
    if ((net <0) || (net > (MAX_CAN_NET-1)))
	return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_ENABLE;
    msg.net     = net;
    msg.id      = swap(id);
    return(canCmmd(&msg));
    }


int canDisableId
    (
    int net,
    int id
    )
    {
    CAN_MSG msg;

    if ((id < 0) || (id > 2047))
	return(-1);
    if ((net <0) || (net > (MAX_CAN_NET-1)))
	return(-1);
    memset(&msg,0,sizeof(msg));
    msg.cmmd    = CAN_HM_DISABLE;
    msg.net     = net;
    msg.id      = swap(id);
    return(canCmmd(&msg));
    }

